/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * Authors: Charles Kerr <charles@skywalker.ecn.ou.edu>
 *          Matt Eagleson <e.Messiah@superpimp.org>
 *          Jason Leach <leach@wam.umd.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <string.h>
#include <ctype.h>
#include "pan.h"
#include "../rx/rxposix.h"

#include "xpm/book_open.xpm"
#include "xpm/book_closed.xpm"
#include "xpm/mini_page.xpm"
#include "xpm/greencheck.xpm"
#include "xpm/bluecheck.xpm"
#include "xpm/x.xpm"
#include "xpm/binary.xpm"
#include "xpm/binary_complete.xpm"
#include "xpm/binary_incomplete.xpm"


/**********************************************************
 * Private Variables
 **********************************************************/

/* search */
static GtkCTreeNode* find_article_next_pos = NULL;
static GtkWidget *find_dialog;

/* article list */
static GnomePixmap *articlelist_closed_pixmap = NULL;
static GnomePixmap *articlelist_open_pixmap = NULL;
static GnomePixmap *decode_queue_pixmap = NULL;
static GnomePixmap *download_queue_pixmap = NULL;
static GnomePixmap *mp_complete_pixmap = NULL;
static GnomePixmap *mp_incomplete_pixmap = NULL;
static GnomePixmap *error_pixmap = NULL;
static GnomePixmap *article_read_pixmap = NULL;
static GnomePixmap *binary_decoded_pixmap = NULL;

/**********************************************************
 * Public Variables
 **********************************************************/

GdkColormap *cmap;
GdkColor new_color;
GdkColor read_color;
GdkColor unread_color;
GdkColor killfile_color;

article_data *current_article = NULL;
GtkWidget *article_ctree_menu;

GtkStyle *killfile_style = NULL;
GtkStyle *unread_style = NULL;
GtkStyle *read_style = NULL;
GtkStyle *new_style = NULL;

/**********************************************************
 * Private Functions
 **********************************************************/

/* search */
static void find_article (const gchar*, const gchar*, gboolean, gboolean );
static void find_dialog_clicked_cb (GnomeDialog *dialog, int button, GtkWidget *find_entry);

/* sorting the ui tree */
static gint articlelist_compare_subjects ( GtkCList*, gconstpointer, gconstpointer );
static gint articlelist_compare_text (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2);
static gint articlelist_compare_commatized (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2); 

/* threading */
static int mp_strcasecmp ( const char *mp, const char *ph );
//static const char* skip_reply_leader ( const char* a );
static int pp_adata_message_id_comp ( const void*, const void* );
static int pp_adata_subject_comp ( const void*, const void* );
static int pp_adata_loc_comp ( const void*, const void* );
static int pp_adata_time_comp ( const void*, const void* );
/* Not currently used
 * static int pp_adata_author_comp ( const void*, const void* );
 */
static int pp_adata_multipart_comp ( const void*, const void* );

/* ui */
static GtkCTreeNode* articlelist_get_node_from_message_id ( const char* message_id, int* index );
static void articlelist_update ( group_data*, article_data**, int qty );
static void build_article_ctree_recursive ( GtkCTree*, GtkCTreeNode*, GSList* );
static GnomePixmap* get_article_pixmap ( const article_data*a );
static void articlelist_button_press (GtkWidget*, GdkEventButton* );
static void articlelist_click_column (GtkCList *, int col_clicked );
static void article_ctree_destroy_cb ( void );
static void articlelist_sort ( void );
static void articlelist_sort_init ( group_data *gdata );
static void articlelist_update_node_fast ( GtkCTreeNode*, const article_data* );
static void thread_articles ( GSList*, int list_length, article_data*** setme_array, int* setme_qty );



/**********************************************************
 * Articlelist popup menu
 **********************************************************/

static GnomeUIInfo articlelist_menu_popup[] = {
	{ GNOME_APP_UI_ITEM, N_("Open in Window"), N_("Open this message for reading."), message_read_window },
	{ GNOME_APP_UI_ITEM, N_("Print"), N_("Print this message out to a physical medium."), not_implemented },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Reply to Sender"), N_("Create a reply to the sender."), message_reply_window },
	{ GNOME_APP_UI_ITEM, N_("Reply to Group"), N_("Post a reply to this message on the news server."), message_reply_window },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Save Attachment"), N_("Save Attachments."), articlelist_selected_decode,
	  NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 'D', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM, N_("Open Attachment"), N_("Open Attachments."), articlelist_selected_open,
	  NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 'O', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM, N_("Delete Message"), N_("Delete message from database."), articlelist_selected_delete },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Mark as Read"), N_("Marks current message as read."), articlelist_selected_mark_read },
	{ GNOME_APP_UI_ITEM, N_("Mark as Unread"), N_("Marks the current message as unread."), articlelist_selected_mark_unread },
	GNOMEUIINFO_SEPARATOR,
        { GNOME_APP_UI_ITEM, N_("Mark All Articles Read"), N_("Marks the entire group as read."), articlelist_all_mark_read },
	GNOMEUIINFO_END
};

typedef struct
{
	article_db db; 
	int total_qty;
	int read_qty;
	unsigned int first;
}   
ExpireCountStruct;

static void
articlelist_expire_func (const char *key, const char *data, gpointer user_data)
{
	ExpireCountStruct* count = (ExpireCountStruct*)user_data;

	if (count->first > atoi (ahdb_field (data, AHDB_FIELD_NUMBER))) {
			if (atoi (ahdb_field (data, AHDB_FIELD_STATE)) & STATE_READ)
				count->read_qty--;
			count->total_qty--;
			ahdb_erase (count->db, NULL, key);
	}
	gui_progress_next_step ();
}

void
articlelist_expire (article_db db, group_data *gdata)
{
	ExpireCountStruct count;

	count.db = db;
	count.total_qty = 0;
	count.read_qty = 0;
	count.first = group_get_attrib_i (gdata, "First");

	ahdb_foreach (db, articlelist_expire_func, &count);

	group_add_attrib_i (gdata, "Read", count.read_qty);
	group_add_attrib_i (gdata, "Total", count.total_qty);
	log_add_va ("%d (%d read) articles expired in group %s", -count.total_qty, -count.read_qty, gdata->name );
}

static const char*
articlelist_get_row_message_id ( int i )
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = gtk_ctree_node_nth (tree, i);
	if ( !node )
		return NULL;
	return (const char*) gtk_ctree_node_get_row_data (tree, node);
}

void
articlelist_set_selected_message_id ( const char* message_id )
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = articlelist_get_node_from_message_id (message_id, NULL);
	GtkCTreeNode* n;

	g_assert ( node!=NULL );

	gtk_ctree_unselect_recursive (tree, gtk_ctree_node_nth(tree, 0));

	for ( n=node; n!=NULL; n=GTK_CTREE_ROW(n)->parent )
		gtk_ctree_expand(tree, n);

	//first_row = ROW_FROM_YPIXEL (clist, 0);
	//last_row = ROW_FROM_YPIXEL (clist, clist->clist_window_height);

	gtk_ctree_select (tree, node);

	if ( !gtk_ctree_node_is_visible (tree, node) )
		gtk_ctree_node_moveto (tree, node, 0, 15, 0);
}

const char*
articlelist_get_selected_message_id ( void )
{
	const char* message_id = NULL;

	const GList* l = GTK_CLIST(Pan.article_ctree)->selection;
	if ( l != NULL )
		message_id = (const char*) gtk_ctree_node_get_row_data ( 
			GTK_CTREE(Pan.article_ctree),
			GTK_CTREE_NODE(l->data) );

	return message_id;
}

static GtkCTreeNode*
articlelist_node_next ( GtkCTreeNode* node )
{
	GtkCTreeNode* n;

	if ( !node )
		return NULL;

	if (GTK_CTREE_ROW(node)->children)
		return GTK_CTREE_ROW(node)->children;

	for (n=node; n!=NULL; n=GTK_CTREE_ROW(n)->parent)
		if (GTK_CTREE_ROW(n)->sibling)
			return GTK_CTREE_ROW(n)->sibling;

	return NULL;
}


static GtkCTreeNode*
articlelist_node_prev ( GtkCTreeNode* node )
{
	GtkCTreeNode *parent;
	GtkCTreeNode *sibling;
	GtkCTreeNode *tmp;
	
	if ( !node )
		return NULL;

	/* get parent */	
	parent = GTK_CTREE_ROW(node)->parent;
	if ( !parent )
		return NULL;

	/* parent's first child */
	sibling=GTK_CTREE_ROW(parent)->children;
	if ( sibling==node ) /* this is the first child */
		return parent;

	/* find the preceeding sibling */
	for ( ;; ) {
		tmp = GTK_CTREE_ROW(sibling)->sibling;
		g_assert ( tmp!=NULL );
		if ( tmp == node ) {
			node = sibling;
			break;
		}
		sibling = tmp;
	}

	/* find the absolutely last child of the older sibling */
	for ( ;; )
	{
		GtkCTreeNode* tmp = GTK_CTREE_ROW(node)->children;
		if ( !tmp ) /* no children */
			return node;
		else
		{
			/* make node point to the last sibling with children,
			   or if none of them have children, just return the last node */
			GtkCTreeNode *children=NULL, *tmp2=NULL;
			for ( ; tmp!=NULL; tmp2=tmp, tmp=GTK_CTREE_ROW(tmp)->sibling )
				if ( GTK_CTREE_ROW(tmp)->children )
					children = GTK_CTREE_ROW(tmp)->children;
			if ( !children )
				return tmp2;
			node = children;
		}
	}

	g_assert_not_reached();
}

const char*
articlelist_get_next_message_id ( const char* message_id, int step )
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode *node=NULL;
	const int row_qty = GTK_CLIST(tree)->rows;

	if ( row_qty<2 )
		return NULL;

	/* get working node */
	node = articlelist_get_node_from_message_id ( message_id, NULL );
	g_assert ( node!=NULL );

	/* Find the requested node... */
	while ( step > 0 ) {
		--step;
		node = articlelist_node_next ( node );
	}
	while ( step < 0 ) {
		++step;
		node = articlelist_node_prev ( node );
	}


	/* get step node */
	return (const char*) gtk_ctree_node_get_row_data ( tree, node );
}


article_data *
articlelist_get_next_article ( void )
{
	/* get current message id */
	const char* message_id = articlelist_get_selected_message_id ( );
	if ( !message_id ) message_id = articlelist_get_row_message_id ( 1 );
	if ( !message_id ) return NULL;

	/* move to next article */
	message_id = articlelist_get_next_message_id ( message_id, 1 );
	if ( message_id != NULL )
	{
		article_db db = ahdb_ref ( Pan.active_server, Pan.active_group );
		article_data *adata = ahdb_get ( db, message_id );
		ahdb_unref ( db );
		return adata;
	}

	return NULL;
}

article_data *
articlelist_get_previous_article ( void )
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);

	/* get current message id */
	const char* message_id = articlelist_get_selected_message_id ( );
	if ( !message_id ) message_id = articlelist_get_row_message_id ( GTK_CLIST(tree)->rows-1 );
	if ( !message_id ) return NULL;

	/* move to next article */
	message_id = articlelist_get_next_message_id ( message_id, -1 );
	if ( message_id != NULL )
	{
		article_db db = ahdb_ref ( Pan.active_server, Pan.active_group );
		article_data *adata = ahdb_get ( db, message_id );
		ahdb_unref ( db );
		return adata;
	}

	return NULL;
}

static void
articlelist_button_press (GtkWidget* widget, GdkEventButton* bevent)
{
	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
					      "button_press_event");

	if (bevent->button == 1)
	{
		/* if in paned mode and one_click_preview set,
		 * a single click shows message 
		 */
		if (one_click_preview && viewmode==GUI_PANED )
		{
			gchar* message_id;
			int row=0, column=0;
			gtk_clist_get_selection_info ( GTK_CLIST(widget), bevent->x, bevent->y, &row, &column );
			message_id = (gchar*) gtk_clist_get_row_data ( GTK_CLIST(Pan.article_ctree), row );
			if ( message_id )
				text_set_from_message_id ( message_id, NNTP_ARTICLE_READ, FALSE );
		}
		
		/* notebook layout uses double clicking */
		if ( bevent->type == GDK_2BUTTON_PRESS )
		{
			gchar* message_id;
			int row=0, column=0;
			gtk_signal_emit_stop_by_name (GTK_OBJECT(widget), "button_press_event");
			gtk_clist_get_selection_info ( GTK_CLIST(widget), bevent->x, bevent->y, &row, &column );
			message_id = (gchar*) gtk_clist_get_row_data ( GTK_CLIST(Pan.article_ctree), row );
			if ( message_id ) {
				text_set_from_message_id ( message_id, NNTP_ARTICLE_READ, FALSE );
				gtk_widget_grab_focus ( text );
			}
		}
	}
	else if ( bevent->button == 3 )
	{
		const gboolean sel = GTK_CLIST(Pan.article_ctree)->selection != NULL;
		const gboolean articles_available = GTK_CLIST(Pan.article_ctree)->rows != 0;

		GList* l = GTK_MENU_SHELL(article_ctree_menu)->children;
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* open */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* print*/
		l = l->next; /* separator */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* reply to sender */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* reply to group */
		l = l->next; /* separator */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* save attachment */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* open attachment */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* delete */
		l = l->next; /* separator */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* mark as read */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel ); /* mark as unread */
		l = l->next; /* separator */
		l = l->next; gtk_widget_set_sensitive ( GTK_WIDGET(l->data), articles_available ); /* mark all read */

		gtk_menu_popup (GTK_MENU (article_ctree_menu), NULL, NULL, NULL,
				NULL, bevent->button, bevent->time);
	}
}


/**********************************************************
 * The following stuff implements the "find" functionality
 **********************************************************/

void
find_dialog_destroy_cb (void)
{
	gtk_widget_destroy(find_dialog);
	find_dialog = NULL;
}




typedef struct
{
	article_db db;

	gboolean compare_author;
	gboolean compare_subject;
	gboolean unread_only;

	regex_t author;
	regex_t subject;
} 
header_compare_data;

static
gboolean
compare_article (
	const article_data* adata,
	const header_compare_data* header )
{
	/* if it ain't an article, it ain't a match */
	if ( !adata )
		return -1;

	/* make sure the unread state (if user specified) matches... */
	if ( header->unread_only && (adata->state&STATE_READ) )
		return -1;

	/* make sure the author (if user specified one) matches.. */
	if ( header->compare_author &&
	     regexec(&header->author,adata->author,0,NULL,0) )
		return -1;

	/* make sure the subject (if user specified one) matches.. */
	if ( header->compare_subject &&
	     regexec(&header->subject,adata->subject,0,NULL,0) )
		return -1;

	/* hooray, a match */
	return 0;
}

static
gint
find_compare_func (
	gconstpointer a,
	gconstpointer b )
{
	gint retval = -1;

	g_assert ( b!=NULL );

	if ( a!=NULL )
	{
		const header_compare_data* header = (const header_compare_data*)b;
		const gchar* message_id = (gchar*)a;
		article_data* adata = ahdb_get ( header->db, message_id );
		retval = compare_article ( adata, header );
		article_free ( adata );
	}

	return retval;
}




static
void
find_article (
	const gchar* author,
	const gchar* subject,
	gboolean case_sensitive,
	gboolean unread_only )
{
	GtkCTreeNode* n;
	GtkCTreeNode *node;
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	int regcomp_flags = REG_EXTENDED|REG_NOSUB|(case_sensitive?0:REG_ICASE);
	char* message_id = NULL;

	header_compare_data header_data;
	header_data.compare_author = author!=NULL && !regcomp ( &header_data.author, author, regcomp_flags );
	header_data.compare_subject = subject!=NULL && !regcomp ( &header_data.subject, subject, regcomp_flags );
	header_data.unread_only = unread_only;
	header_data.db = ahdb_ref ( Pan.active_server, Pan.active_group );

	node = gtk_ctree_find_by_row_data_custom (
		tree,
		find_article_next_pos,
		&header_data, 
		(GCompareFunc)find_compare_func );

	ahdb_unref ( header_data.db );

	if ( !node ) {
		find_article_next_pos = NULL;
		return;
	}

	/* select the node.. */
	gtk_clist_freeze (GTK_CLIST(tree));
	gtk_ctree_unselect_recursive (tree, NULL);
	gtk_ctree_select (tree, node);
	for (n = GTK_CTREE_ROW(node)->parent; n; n = GTK_CTREE_ROW(n)->parent)
		gtk_ctree_expand(tree, n);
	gtk_ctree_node_moveto (tree, node, 0, 0, 0);
	gtk_clist_thaw (GTK_CLIST(tree));
	message_id = gtk_ctree_node_get_row_data(tree, node);
	if ( message_id != NULL  )
		text_set_from_message_id ( message_id, NNTP_ARTICLE_READ, FALSE );

	/* Find next node, paying attention to recursion */
	if (GTK_CTREE_ROW(node)->children) {
		find_article_next_pos = GTK_CTREE_ROW(node)->children;
	} else {
		for (n=node; n!=NULL; n=GTK_CTREE_ROW(n)->parent) {
			if (GTK_CTREE_ROW(n)->sibling) {
				find_article_next_pos = GTK_CTREE_ROW(n)->sibling;
				break;
			}
		}
	}

	/* cleanup */
	if ( header_data.compare_author )
		regfree ( &header_data.author );
	if ( header_data.compare_subject )
		regfree ( &header_data.subject );
}

static gchar* find_article__author = NULL;
static gchar* find_article__subject = NULL;
static gboolean find_article__case_sensitive = FALSE;
static gboolean find_article__unread_only = FALSE;

static void
find_dialog_clicked_cb (
	GnomeDialog* dialog,
	int button,
	GtkWidget* table )
{
 	g_return_if_fail ( button==0 || button==1 );
	g_return_if_fail ( table!=NULL );

	if ( button==1 ) /* close button */
	{
		gtk_widget_unmap ( GTK_WIDGET(find_dialog) );
	}
	else if ( button==0 ) /* "find next" button */
	{
		GtkWidget* author_entry;
		GtkWidget* subject_entry;
		GtkWidget* case_sensitive_checkbox;
		GtkWidget* unread_only_checkbox;
		gchar* author = 0;
		gchar* subject = 0;
		gboolean case_sensitive;
		gboolean unread_only;

		/* get author to search for, or NULL if none... */
		author_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "author" );
		g_return_if_fail ( author_entry!=NULL );
		author = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(author_entry) ) );
		g_strstrip ( author );
		if ( !*author ) {
			g_free ( author );
			author = NULL;
		}

		/* get subject to search for, or NULL if none... */
		subject_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "subject" );
		g_return_if_fail ( subject_entry!=NULL );
		subject = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(subject_entry) ) );
		g_strstrip ( subject );
		if ( !*subject ) {
			g_free ( subject );
			subject = NULL;
		}

		/* get case sensitivity... */
		case_sensitive_checkbox = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "case" );
		g_return_if_fail ( case_sensitive_checkbox!=NULL );
		case_sensitive = gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbox) );

		/* get case sensitivity... */
		unread_only_checkbox = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "unread" );
		g_return_if_fail ( unread_only_checkbox!=NULL );
		unread_only = gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(unread_only_checkbox) );

		/* update the find_article fields with these widget values */
		g_free ( find_article__author );
		find_article__author = author;
		g_free ( find_article__subject );
		find_article__subject = subject;
		find_article__case_sensitive = case_sensitive;
		find_article__unread_only = unread_only;

		articlelist_find_next_cb ( );
	}
}

void
articlelist_find_next_cb (void)
{
	find_article (
		find_article__author,
		find_article__subject,
		find_article__case_sensitive,
		find_article__unread_only );
}


void
articlelist_find_text_cb (void)
{
	GtkWidget *table;
	GtkWidget *author_label, *author_entry;
	GtkWidget *subject_label, *subject_entry;
	GtkWidget *case_sensitive_checkbutton;
	GtkWidget *unread_only_checkbutton;

	/* There can be only one! (find window) */
	if (find_dialog) {
		if ( !GTK_WIDGET_MAPPED(GTK_WIDGET(find_dialog)) )
			gtk_widget_map ( GTK_WIDGET(find_dialog) );
		gdk_window_raise (GTK_WIDGET(find_dialog)->window);
		return;
	}
	
	find_dialog = gnome_dialog_new ( _("Find Message"),
					 GNOME_STOCK_PIXMAP_FORWARD,
					 GNOME_STOCK_PIXMAP_CLOSE, NULL);
	gnome_dialog_set_default ( GNOME_DIALOG(find_dialog), 0 );

	table = gtk_table_new ( 4, 2, FALSE );
//	gtk_container_set_border_width ( GTK_CONTAINER(table), GNOME_PAD_SMALL );
//        gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
//        gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	subject_label = gtk_label_new ( _("Subject:") );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_label,
			   0, 1,
			   0, 1,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	subject_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "subject", subject_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_entry,
			   1, 2,
			   0, 1,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_label = gtk_label_new ( _("Author:") );
	gtk_table_attach ( GTK_TABLE(table),
			   author_label,
			   0, 1,
			   1, 2,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "author", author_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   author_entry,
			   1, 2,
			   1, 2,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	unread_only_checkbutton = gtk_check_button_new_with_label ( "Unread Only" );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(unread_only_checkbutton), FALSE );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "unread", unread_only_checkbutton );
	gtk_table_attach ( GTK_TABLE(table),
			   unread_only_checkbutton,
			   0, 2,
			   2, 3,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	case_sensitive_checkbutton = gtk_check_button_new_with_label ( "Case Sensitive" );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbutton), FALSE );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "case", case_sensitive_checkbutton );
	gtk_table_attach ( GTK_TABLE(table),
			   case_sensitive_checkbutton,
			   0, 2,
			   3, 4,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	
//	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(find_dialog)->vbox), table );
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (find_dialog)->vbox), table, FALSE, FALSE, 0);
	
	gtk_signal_connect (GTK_OBJECT(find_dialog), "clicked",
			    GTK_SIGNAL_FUNC(find_dialog_clicked_cb),
			    table);
	gtk_signal_connect (GTK_OBJECT(find_dialog), "destroy",
			    GTK_SIGNAL_FUNC(find_dialog_destroy_cb), NULL);

	gtk_widget_grab_focus (subject_entry);
	gui_popup_draw (find_dialog);
}


/*--------------------------------------------------------------------
 * clear the ctree of all messages
 *--------------------------------------------------------------------*/
void
articlelist_clear (void)
{
	pan_lock();
	gtk_clist_freeze (GTK_CLIST(Pan.article_ctree));
	gtk_clist_clear (GTK_CLIST(Pan.article_ctree));
	gtk_clist_thaw (GTK_CLIST(Pan.article_ctree));
	pan_unlock();
}


/*--------------------------------------------------------------------
 * free an article list (and all threads)
 *--------------------------------------------------------------------*/
void
articlelist_free (GSList *alist)
{
	GSList *p;
	article_data *adata;

	for (p = alist; p != NULL; p = p->next)
	{
		adata = p->data;

		if (adata->threads)
			articlelist_free (adata->threads);
		article_free (adata);
		p->data = NULL;
	}

	g_slist_free (alist);
}


/*--------------------------------------------------------------------
 * load article listing
 *--------------------------------------------------------------------*/
void
articlelist_load (
	group_data* gdata )
{
	article_db db;
	GSList *alist;
	int total = group_get_attrib_i ( gdata, "Total" );
	int loaded_total;

	/* argument sanity check */
	g_return_if_fail ( gdata != NULL );

	/* Clear the currently displayed articlelist, if any */
	if ( Pan.active_group )
		articlelist_clear ();

	/* step 1: load article headers */
        gui_set_status ( _("Loading article headers") );
	gui_progress_init_steps ( total * 4 );
	db = ahdb_ref ( Pan.active_server, gdata );
	alist = ahdb_get_all ( db, gui_progress_next_step_gfunc, NULL );
	ahdb_unref ( db );

	/* check totals */
	loaded_total = g_slist_length ( alist );
	if ( total != loaded_total ) {
		g_warning (
			_("Pan thought there were %d articles; instead we got %d"),
			total,
			loaded_total );
		group_set_attrib_i ( gdata, "Total", total=loaded_total );
	}

	if ( !alist )
	{
		pan_lock ();
		choose_download_dialog (gdata);
		pan_unlock ();
	} 
	else
	{
		log_add_va (
			_("Loaded %d article headers for group %s"),
			total,
			gdata->name );
		articlelist_set_contents ( gdata, alist, total );
	}
}

static GSList*
killfile_articles (group_data* gdata,  GSList* alist, int* alist_total)
{
	GSList* prev = NULL;
	GSList* cur = alist;

	while ( cur != NULL )
	{
		article_data* adata = (article_data*) cur->data;

		const MessageFilterAction action =
			message_filter_get_action ( gdata, adata );

		if ( action & MESSAGE_FILTER_ACTION_DELETE )
		{
			/* mark for deletion */
			GSList* del1me = cur;

			/* unlink */
			if ( prev )
				prev->next = cur->next;
			else
				alist = cur->next;
			cur = cur->next;

			/* delete article from list */
			--*alist_total;
			g_slist_free_1 ( del1me );
			article_free ( adata );
			del1me->data = NULL;
		}
		else
		{
			if ( action & MESSAGE_FILTER_ACTION_COLOR )
				adata->state |= STATE_KILLFILE;

			prev = cur;
			cur = cur->next;
		}

		gui_progress_next_step ( );
	}

	return alist;
}


void
articlelist_set_contents (
	group_data* gdata,
	GSList* alist,
	int alist_total )
{
	article_data** buf = NULL;
	int buf_qty = 0;

	gui_set_status ( _("Applying Killfile") );
	alist = killfile_articles ( gdata, alist, &alist_total );

	articlelist_sort_init ( gdata );
	gui_set_status (_("Threading article list"));
	thread_articles ( alist, alist_total, &buf, &buf_qty );

	gui_set_status (_("Displaying article list"));
	articlelist_update ( gdata, buf, buf_qty );

//	articlelist_free ( alist );
}




//============================================================================

typedef struct
{
	GtkCTreeNode* node;
	article_db db;
	int total_qty;
	int read_qty;
	GFunc func;
}
CountStruct;

static gboolean
ancestor_is_selected ( const GtkCTree* tree, const GtkCTreeNode* node )
{
	const GList* list =  GTK_CLIST(tree)->selection;
	if ( !list )
		return FALSE;

	for ( ;; ) {
		node = GTK_CTREE_ROW(node)->parent;
		if ( !node )
			return FALSE;
		if ( g_list_index ( (GList*)list, (gpointer)node ) != -1 )
			return TRUE;
	}

        g_assert_not_reached ( );
}

static
void
articlelist_db_erase_article_gfunc ( gpointer data,
				     gpointer user_data )
{
	article_data* adata = (article_data*)data;
	CountStruct* count = (CountStruct*)user_data;
	ahdb_erase ( count->db, adata, NULL);

	if ( count->node == find_article_next_pos )
		find_article_next_pos = NULL;
}

static
void
articlelist_db_mark_unread_gfunc ( gpointer data,
				   gpointer user_data )
{
	article_data* adata = (article_data*)data;
	if ( adata->state & STATE_READ )
	{
		CountStruct* count = (CountStruct*)user_data;
		adata->state &= ~STATE_READ;
		ahdb_save ( count->db, adata, FALSE );
		articlelist_update_node_fast ( count->node, adata );
	}
}


static
void
articlelist_db_mark_read_gfunc ( gpointer data,
				 gpointer user_data )
{
	article_data* adata = (article_data*)data;
	if ( ! ( adata->state & STATE_READ ) )
	{
		CountStruct* count = (CountStruct*)user_data;
		adata->state |= STATE_READ;
		ahdb_save ( count->db, adata, FALSE );
		articlelist_update_node_fast ( count->node, adata );
	}
}


static
void
count_articles ( GtkCTree* ctree,
		 GtkCTreeNode* node,
		 gpointer data )
{
	CountStruct* count = (CountStruct*)data;
	const gchar* message_id = NULL;
	article_data* a = NULL;

	/* sanity checking */
	g_return_if_fail ( count!=NULL );
	g_return_if_fail ( ctree!=NULL );
	g_return_if_fail ( node!=NULL );

	/* get the article */
	message_id = (gchar*) gtk_ctree_node_get_row_data ( ctree, node );
	if ( !message_id ) {
		if ( node == gtk_ctree_node_nth(ctree,0) ) /* name-of-group */
			return;
		else g_assert ( message_id!=NULL );
	}
	a = ahdb_get ( count->db, message_id );
	g_return_if_fail ( a!=NULL );

	/* tally up the read/total */
	++count->total_qty;
	if ( a->state & STATE_READ )
		++count->read_qty;

	/* call the client-specified function */
	count->node = node;
	(*count->func)(a, data);

	/* cleanup */
	article_free ( a );
}

void
articlelist_selected_delete ( void )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	server_data* server = Pan.active_server;
	group_data* group = Pan.active_group;
	GSList* l = NULL;
	GSList* prune_branches = NULL;
	CountStruct count;

	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref ( server, group );
	count.func = articlelist_db_erase_article_gfunc;

	gtk_clist_freeze ( GTK_CLIST(tree) );

	/* first we update the article header db... */
	for ( l=(GSList*)GTK_CLIST(tree)->selection; l!=NULL; l=l->next ) {
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) ) {
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
			prune_branches = g_slist_prepend ( prune_branches, l->data );
		}
	}

	/* then we update the group attributes db... */
	group_add_attrib_i ( group, "Read", -count.read_qty );
	group_add_attrib_i ( group, "Total", -count.total_qty );

	/* then finally we update the ui. */
	gtk_ctree_unselect_recursive ( tree, NULL );
	for ( l=prune_branches; l!=NULL; l=l->next )
		gtk_ctree_remove_node ( tree, GTK_CTREE_NODE(l->data) );

	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* cleanup */
	g_slist_free ( prune_branches );
	ahdb_sync ( count.db );
	ahdb_unref ( count.db );
}

static
void
articlelist_mark_read (
	GSList* l )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	group_data* group = Pan.active_group;
	CountStruct count;
	int newly_read;

	/* argument sanity checking */
	g_return_if_fail ( l!=NULL );

	/* setup */
	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref ( Pan.active_server, group );
	count.func = articlelist_db_mark_read_gfunc;

	/* mark them all as read */
	gtk_clist_freeze ( GTK_CLIST(tree) );
	for ( ; l!=NULL; l=l->next )
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) )
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* update the group attributes */
	newly_read = count.total_qty - count.read_qty;
	group_add_attrib_i ( group, "Read", newly_read );

	/* cleanup */
	ahdb_sync ( count.db );
	ahdb_unref ( count.db );
}

void
articlelist_selected_mark_read ( void )
{
	articlelist_mark_read ( (GSList*) GTK_CLIST(Pan.article_ctree)->selection );
}

void
articlelist_all_mark_read ( void )
{
	GSList* l = g_slist_prepend ( NULL, gtk_ctree_node_nth ( GTK_CTREE(Pan.article_ctree), 0 ) );
	articlelist_mark_read ( l );
	g_slist_free ( l );
}

void
articlelist_selected_mark_unread ( )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	server_data* server = Pan.active_server;
	group_data* group = Pan.active_group;
	GSList* l;
	CountStruct count;

	/* setup */
	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref ( server, group );
	count.func = articlelist_db_mark_unread_gfunc;

	/* mark them all as read */
	gtk_clist_freeze ( GTK_CLIST(tree) );
	for ( l=(GSList*)GTK_LIST(tree)->selection; l!=NULL; l=l->next )
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) )
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* update the group attributes */
	group_add_attrib_i ( group, "Read", -count.read_qty );

	/* cleanup */
	ahdb_unref ( count.db );
}

static void
articlelist_selected_decode_impl (guint flags)
{
	server_data* sdata = Pan.active_server;
	group_data* gdata = Pan.active_group;
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );

	/* loop through all the selected articles */
	GSList* l;
	article_db db = ahdb_ref ( sdata, gdata );
	gtk_clist_freeze ( GTK_CLIST(tree) );
	for ( l=(GSList*)GTK_CLIST(tree)->selection; l!=NULL; l=l->next )
	{
		article_data* adata = NULL;

		/* build an array of this node and all its children. */
		GtkCTreeNode* node = GTK_CTREE_NODE(l->data);
		GSList* list = g_slist_append ( NULL, gtk_ctree_node_get_row_data ( tree, node ) );
		
		if (list->data) {
			adata = ahdb_get ( db, (gchar*)list->data );

			if (adata->parts > 1) {
				for ( node=GTK_CTREE_ROW(node)->children; node!=NULL; node=GTK_CTREE_ROW(node)->sibling )
					list = g_slist_append ( list, gtk_ctree_node_get_row_data ( tree, node ) );
			}

			/* queue for decode */
			queue_add_article_decode ( sdata, gdata, list, flags );
		
			/* update the list to show this is queued */
			adata->state |= STATE_DECODE_QUEUED;
			articlelist_update_node_fast ( GTK_CTREE_NODE(l->data), adata );
	
			/* cleanup this loop */
			article_free ( adata );
		
		}
		g_slist_free ( list );
	}

	/* cleanup */
	ahdb_unref ( db );
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );
}
	
void
articlelist_selected_decode ( void )
{		
	articlelist_selected_decode_impl ( QUEUE_ARTICLE_DECODE|QUEUE_BACKGROUND );
}

void
articlelist_selected_open ( void )
{
	articlelist_selected_decode_impl ( QUEUE_ARTICLE_DECODE|QUEUE_BACKGROUND|QUEUE_OPEN );
}

void
articlelist_select_all ( void )
{
	gtk_ctree_select_recursive ( GTK_CTREE(Pan.article_ctree), NULL );
}


//============================================================================


static
GnomePixmap*
get_article_pixmap (
	const article_data* adata )
{
	g_return_val_if_fail ( adata!=NULL, NULL );

	if (adata->state & STATE_ERROR)
		return error_pixmap;

	if (adata->state & STATE_DECODED)
		return binary_decoded_pixmap;

	if (adata->state & STATE_READ)
		return article_read_pixmap;

	if (adata->state & STATE_DOWNLOAD_QUEUED)
		return download_queue_pixmap;

	if (adata->state & STATE_DECODE_QUEUED)
		return decode_queue_pixmap;

	if (adata->state & STATE_MULTIPART_ALL)
		return mp_complete_pixmap;

	if (adata->state & STATE_MULTIPART_SOME)
		return mp_incomplete_pixmap;

	return NULL;
}


static
void
articlelist_set_node_style (
	GtkCTreeNode* node,
	const article_data* adata )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );

	if ( !unread_style )
	{
		GdkFont* font = NULL;
		if ( articlelist_font != NULL )
			font = gdk_font_load ( articlelist_font );

		killfile_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		killfile_style->fg[0] = unread_color;

		unread_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		unread_style->fg[0] = unread_color;

		read_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		read_style->fg[0] = read_color;

		new_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		new_style->fg[0] = new_color;

		if ( font ) {
			killfile_style->font = font;
			unread_style->font = font;
			read_style->font = font;
			new_style->font = font;
		}
	}

	if ( articlelist_font )
	{
		if ( adata->state & STATE_KILLFILE )
			gtk_ctree_node_set_row_style (tree, node, killfile_style);
		else if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_row_style (tree, node, read_style);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_row_style (tree, node, new_style);
		else
			gtk_ctree_node_set_row_style (tree, node, unread_style);
	}	
	else /* If the font isn't specified by the user, just set the color */
	{
		if ( adata->state & STATE_KILLFILE )
			gtk_ctree_node_set_foreground (tree, node, &killfile_color);
		else if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_foreground (tree, node, &read_color);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_foreground (tree, node, &new_color);
		else
			gtk_ctree_node_set_foreground (tree, node, &unread_color);
	}
}

static
GtkCTreeNode*
add_article_to_ctree (
	GtkCTree* tree,
	GtkCTreeNode* parent,
	const article_data* adata)
{
	char *text[4];
	char line_count_buf[16];
	static int n = 0;
	GtkCTreeNode *node;
	GnomePixmap* icon = get_article_pixmap ( adata );

        /* this is sortable with a simple strcmp */
        char timebuf[32];
        const struct tm local_tm = *localtime ( &adata->date );
        strftime ( timebuf, sizeof(timebuf), "%Y/%m/%d %H:%M", &local_tm );

	/* line count */
	{
		gulong lcountsum = adata->linecount;
		GSList *l;
		for ( l = adata->threads; l; l = l->next)
			lcountsum += ((article_data *)l->data)->linecount;
                commatize_ulong ( lcountsum, line_count_buf );
	}

	text[0] = adata->subject;
	text[1] = line_count_buf;
	text[2] = timebuf;
	text[3] = adata->author;

	gui_progress_next_step ( );

	node = gtk_ctree_insert_node (
		tree, parent, NULL,
		text, 5,
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		FALSE, FALSE );

	articlelist_set_node_style (node, adata);

	gtk_ctree_node_set_row_data_full (tree,
					  node,
					  g_strdup(adata->message_id),
					  g_free);

	if ( ++n == 250 ) {
		n = 0;
		pan_unlock();
		pan_lock();
	}

	return node;
}

static void
build_article_ctree_recursive (
	GtkCTree* tree,
	GtkCTreeNode* parent,
	GSList* list )
{
	GSList *pos;
        for ( pos=list; pos!=NULL; pos=pos->next )
        {
		article_data* adata = (article_data *) pos->data;
		GtkCTreeNode* node = add_article_to_ctree (tree, parent, adata);
		if ( adata->threads!=NULL ) //&& !( data->state & STATE_MULTIPART_ALL ) )
			build_article_ctree_recursive (tree, node, adata->threads);
        }
}


static void
articlelist_update (
	group_data* gdata,
	article_data** article_buf,
	int article_qty )
{
	char *text[4];
	int i;
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	GtkCTreeNode *root;

	g_assert ( gdata != NULL );

	text[0] = gdata->name;
	text[1] = text[2] = text[3] = NULL;

	gui_set_status (_("Updating article display"));
//	gui_progress_init_steps ( articles_qty );

	articlelist_clear ();
	find_article_next_pos = NULL;

	//gtk_widget_unmap (GTK_WIDGET (Pan.article_ctree));

	pan_lock ( );
	gtk_clist_freeze ( GTK_CLIST(tree) );
	/* add root node */
	root = gtk_ctree_insert_node (
		tree, NULL, NULL,
		text, 5, articlelist_closed_pixmap->pixmap,
		articlelist_closed_pixmap->mask,
		articlelist_open_pixmap->pixmap,
		articlelist_open_pixmap->mask, FALSE, TRUE );
	gtk_ctree_node_set_row_data (tree, root, NULL);

	/* add top-level articles */
	for ( i=0; i!=article_qty; ++i ) {
		article_data* adata = article_buf[i];
		GtkCTreeNode* node = add_article_to_ctree ( tree, root, adata );
		if ( adata->threads!=NULL )
			build_article_ctree_recursive ( tree, node, adata->threads );
	}

	gtk_clist_thaw ( GTK_CLIST(tree) );
	//gtk_widget_map (GTK_WIDGET (Pan.article_ctree));
	gtk_widget_queue_resize ( Pan.article_ctree );
	pan_unlock ( );
}


static
void 
articlelist_update_node_fast (
	GtkCTreeNode* node,
	const article_data* adata )
{
	GnomePixmap* icon = get_article_pixmap ( adata );
	gtk_ctree_set_node_info (
		GTK_CTREE(Pan.article_ctree), node,
		adata->subject, 5,
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		GTK_CTREE_ROW(node)->is_leaf,
		GTK_CTREE_ROW(node)->expanded );

	articlelist_set_node_style ( node, adata );
}


static GtkCTreeNode*
articlelist_get_node_from_message_id ( const char* message_id, int* index )
{
        GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = gtk_ctree_node_nth(tree, 0);
	const char* pch;
	int i=0;

	g_assert ( message_id != NULL );

	for ( ;; )
	{
		++i;
		node = articlelist_node_next ( node );
		if ( !node )
			return NULL;

		pch = (const char*) gtk_ctree_node_get_row_data(tree, node);
		if ( !strcmp ( message_id, pch ) ) {
			if ( index )
				*index = i;
			return node;
		}
	}

	g_assert_not_reached ( );
}

void
articlelist_update_node (
	const article_data* adata )
{
	GtkCTreeNode* node = articlelist_get_node_from_message_id ( adata->message_id, NULL );
        if ( node != NULL )
                articlelist_update_node_fast ( node, adata );
}


/*********************************************************************
 * The following stuff implements the "thread articles" functionality
 ********************************************************************/

static   
int 
mp_strcasecmp (
	register const char* mp,
	register const char* ph )
{
	register int val;

	while ( *mp && *ph )
	{
		val = *mp - *ph;
		if ( val ) {
			val = toupper(*mp) - toupper(*ph);
			if ( val )
				return val;
		}

		if ( *mp=='(' || *mp=='[' )
		{
			++mp;
			++ph;
			if ( isdigit((int)*mp) && isdigit((int)*ph) )
			{
				while ( isdigit((int)*mp) ) ++mp;
				while ( isdigit((int)*ph) ) ++ph;
			}
		}
		else
		{
			++mp;
			++ph;
		}
	}

	return 0;
}


/**
 * Skip the "Re: " part of a subject header, if any
 * @param subject
 * @return the non-"Re:" portion of the subject header
 */
#define skip_reply_leader(a) ((a!=NULL && toupper(a[0])=='R' && toupper(a[1])=='E' && toupper(a[2])==':')?a+4:a)
#if 0
static
const char*
skip_reply_leader ( register const char* a )
{
	if ( !a )
		return NULL;

	if ( toupper(a[0])=='R' && toupper(a[1])=='E' && toupper(a[2])==':' )
		return a+4;

	return a;
}
#endif

/* 
 * not currently used
static int
pp_adata_author_comp (
	const void* va,
	const void* vb )
{
        const article_data* a = *(const article_data**)va;
        const article_data* b = *(const article_data**)vb;
	return strcmp ( a->author, b->author );
}
*/

static int
pp_adata_loc_comp (
	const void* va,
	const void* vb )
{
        const article_data* a = *(const article_data**)va;
        const article_data* b = *(const article_data**)vb;
	if ( !a->threads && !b->threads )
		return a->linecount - b->linecount;
	else {
		gulong a_loc = a->linecount;
		gulong b_loc = b->linecount;
		GSList *l;
		for ( l=a->threads; l; l=l->next) a_loc += ((article_data *)l->data)->linecount;
		for ( l=b->threads; l; l=l->next) b_loc += ((article_data *)l->data)->linecount;
		return a_loc - b_loc;
	}
	g_assert_not_reached ( );
} 

static int
pp_adata_time_comp (
	const void* va,
	const void* vb )
{
        const article_data* a = *(const article_data**)va;
        const article_data* b = *(const article_data**)vb;
	return a->date - b->date;
}

/**
 * Sort this way:
 * (1) by subject, with replies coming after non-replies of the same subject, but before other subjects.
 * (2) by time
 */
static int
pp_adata_subject_comp (
	const void* va,
	const void* vb )
{
        const article_data* a = *(const article_data**)va;
        const char* subj_a = skip_reply_leader ( a->subject );
        const gboolean a_is_re = a->subject != subj_a;

        const article_data* b = *(const article_data**)vb;
        const char* subj_b = skip_reply_leader ( b->subject );
        const gboolean b_is_re = b->subject != subj_b;

        /* if they're not related, a simple strcmp does the job */
        int value = g_strcasecmp ( subj_a, subj_b );
	if ( value )
		return value;

	/* if one but not both is a reply, the reply goes second */
	if ( !!a_is_re != !!b_is_re )
		return a_is_re ? 1 : -1;

	/* oldest goes first... */
	value = a->date - b->date;
	if ( value )
		return value;

	return 0;
}

/**
 * Sort by message id
 */
static int
pp_adata_message_id_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = *(const article_data**)va;
	const article_data* b = *(const article_data**)vb;
	return strcmp ( a->message_id, b->message_id );
}


/**
 * Works like article_subject_cmp, but handles multiparts better
 */
static int
p_adata_mp_subject_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = (const article_data*) va;
	const char* subj_a = skip_reply_leader ( a->subject );
	const gboolean a_is_re = a->subject != subj_a;

	const article_data* b = (const article_data*) vb;
	const char* subj_b = skip_reply_leader ( b->subject );
	const gboolean b_is_re = b->subject != subj_b;

	/* get the easy stuff out of the way */
	int value = mp_strcasecmp ( subj_a, subj_b );
	if ( value )
		return value;

	/* check multipart */
	value = a->part - b->part;
	if ( value )
		return value;

	/* if one but not both is a reply, the reply goes second */
	if ( a_is_re != b_is_re )
		return a_is_re ? 1 : -1;

	/* oldest goes first... */
	value = a->date - b->date;
	if ( value )
		return value;

	return 0;
}

/**
 * Sort this way:
 * (1) by subject, with replies after non-replies of same subject,
 *     but before other subjects.
 * (2) by time
 */
#if 0
static int
pp_adata_subject_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = *(const article_data**)va;
	const char* subj_a = skip_reply_leader ( a->subject );
	const gboolean a_is_re = a->subject != subj_a;

	const article_data* b = *(const article_data**)vb;
	const char* subj_b = skip_reply_leader ( b->subject );
	const gboolean b_is_re = b->subject != subj_b;

	/* if they're not related, a simple strcmp does the job */
	int value = g_strcasecmp ( subj_a, subj_b );
	if ( value )
		return value;

	/* if one but not both is a reply, the reply goes second */
	if ( !!a_is_re != !!b_is_re )
		return a_is_re ? 1 : -1;

	/* oldest goes first... */
	value = a->date - b->date;
	if ( value )
		return value;

	return 0;
}
#endif

/**
 * Sort thusly:
 * (1) in ascending number of part qty
 * (2) by subject via mp_strcasecmp
 * (3) by part number
 * This should lump all the multiparts together in ascending order.
 */
static int
pp_adata_multipart_comp (
	const void* va,
	const void* vb )
{
	register const article_data* a = *(const article_data**)va;
	register const article_data* b = *(const article_data**)vb;

	/* parts */
	int val = a->parts - b->parts;
	if ( val )
		return val;

	/* subject */
	val = *a->subject - *b->subject;
	if ( val )
		return val;
	val = mp_strcasecmp ( a->subject, b->subject );
	if ( val )
		return val;

	/* part number */
	val = a->part - b->part;
	if ( val )
		return val;

	/* did someone repost this? */
	val = a->date - b->date;
	if ( val )
		return val;

	/* they are equal */
	return 0;
}

typedef int (*compare_func)( const void*, const void*);


/**
 * Thread the articles specified in list
 */
static void
thread_articles (
	GSList* list,
	int list_len,
	article_data*** setme_array,
	int* setme_qty )
{
	GSList* p;
	int i, j;
	article_data **refs, **subjs, **parts, **buf;
	article_data search_a, *p_search_a=&search_a;
	int remove_qty = 0;
	int* remove_from_top = NULL;
	int article_qty = list_len;
        const int sort_column = GTK_CLIST(Pan.article_ctree)->sort_column;
	compare_func func = NULL;

	g_assert ( list!=NULL );

	/* if nothing to do, do nothing */
	if ( !article_qty ) {
		*setme_array = 0;
		*setme_qty = 0;
	}

	/* make a reference copy of the articles... */
	buf = g_new ( article_data*, article_qty );
	for ( i=0, p=list; p!=NULL; p=p->next, ++i ) {
		buf[i] = (article_data*) p->data;
		g_assert ( buf[i] != NULL );
	}
	g_slist_free ( list );
	list = NULL;
	g_assert ( i==article_qty );

	/* make various sorted copies of the articles for fast searching */
	refs = g_memdup ( buf, sizeof(article_data*)*article_qty );
	subjs = g_memdup ( buf, sizeof(article_data*)*article_qty );
	parts = g_memdup ( buf, sizeof(article_data*)*article_qty );
	qsort ( refs, article_qty, sizeof(article_data*), pp_adata_message_id_comp );
	qsort ( subjs, article_qty, sizeof(article_data*), pp_adata_subject_comp );
	qsort ( parts, article_qty, sizeof(article_data*), pp_adata_multipart_comp );

	/* housekeeping */
	remove_qty = 0;
	remove_from_top = g_new ( int, article_qty );
//	gui_progress_init_steps ( article_qty );

	/* thread the articles */
	for ( i=0; i!=article_qty; ++i )
	{
		article_data *parent = NULL;
		article_data *a = buf[i];
		int index = -1;

		/* visual feedback */
		gui_progress_next_step ( );

		/* thread by reference
		   (except for multiparts, which are top-level: !a->parts ) */
		if ( a->references && *a->references=='<' && !a->parts )
		{
			gboolean exact = FALSE;
			search_a.message_id = strrchr ( a->references, '<' );
			index = lower_bound ( &p_search_a, refs, article_qty, sizeof(article_data*), pp_adata_message_id_comp, &exact );
			if ( exact )
				parent = refs[index];
		}

		/* thread by multipart */
		if ( parent==NULL && a->parts>1 && a->part>1 )
		{
			search_a.subject = a->subject;
			search_a.part = 1;
			search_a.parts = a->parts;
			search_a.date = 0; /* unlikely to get exact match :) */
			index = lower_bound ( &p_search_a, parts, article_qty, sizeof(article_data*), pp_adata_multipart_comp, NULL );

			if ( index<article_qty && parts[index]!=buf[i] )
				parent = parts[index];
		}

		/* thread by subject */
		if ( parent==NULL && skip_reply_leader(a->subject)!=a->subject )
		{
			search_a.subject = (char*) skip_reply_leader ( a->subject );
			search_a.date = 0; /* unlikely to get exact match :) */
			index = lower_bound ( &p_search_a, subjs, article_qty, sizeof(article_data*), pp_adata_subject_comp, NULL );
			if ( index<article_qty )
			{
				article_data* b = subjs[index];
				if ( !g_strcasecmp ( search_a.subject, b->subject ) ) /* simple case of source/reply */
					parent = b;
				else if ( !g_strcasecmp ( a->subject, b->subject ) && b->date<a->date ) /* 2 replies, no source: thread under oldest */
					parent = b;
			}
		}

		/* this article has a parent */
		if ( parent!=NULL )
		{
			parent->threads = g_slist_insert_sorted ( parent->threads, a, p_adata_mp_subject_comp );
			remove_from_top[remove_qty++] = i;
		}
	}

	g_free ( parts );
	g_free ( subjs );

	/* NULL out the elements to be removed from top level */
	for ( i=0; i!=remove_qty; ++i )
		buf[remove_from_top[i]] = NULL;
	g_free ( remove_from_top );

	/* remove the NULL elements.
	   We recycle refs here for convenience */
	for ( i=j=0; i!=article_qty; ++i )
		if ( buf[i]!=NULL )
			refs[j++] = buf[i];
	article_qty = j;
	g_free ( buf );
	buf = refs;
	refs = NULL;

	/* set the multipart state (all/partial) */
	for ( i=0; i!=article_qty; ++i )
	{
		GSList* p;

		article_data* a = buf[i];
		if ( !a->parts )
			continue;
		if ( a->part==0 ) /* this is treated separately, so don't give it a multipart state */
			continue;
		if ( a->part > 1 ) { /* if a top-level's part is >1, we already know it's incomplete */
			a->state |= STATE_MULTIPART_SOME;
			continue;
		}
		if ( a->part==1 && a->parts==1 ) { /* handle the single-part attachment message */
			a->state |= STATE_MULTIPART_ALL;
			continue;
		}

		/* make sure we have each multipart. */
		for ( j=a->part+1, p=a->threads; j!=a->parts && p!=NULL; p=p->next )
		{
			article_data* b = (article_data*)(p->data);
			if ( b->part > j )
				break;
			else if ( b->part == j )
				++j;
			else
				; /* a repost of a multipart section? */
		}
		if ( j==a->parts ) {
			a->state |= STATE_MULTIPART_ALL;
		} else {
			a->state |= STATE_MULTIPART_SOME;
		}
	}

	/* sort the resulting list based on the sort column */
	switch ( sort_column ) {
		case 0: func = pp_adata_subject_comp; break;
		case 1: func = pp_adata_loc_comp; break;
		case 2: func = pp_adata_time_comp; break;
		case 3: func = pp_adata_time_comp; break;
		default: g_assert_not_reached();
	}
	qsort ( buf, article_qty, sizeof(article_data*), func );

	/* return the result */
	*setme_array = buf;
	*setme_qty = article_qty;
}


/***************************************************************
 * The following stuff implements the tree sorting functionality
 **************************************************************/

static
void
articlelist_sort_init (
	group_data *gdata)
{
	gint sort_style = 0;

	sort_style = group_get_attrib_i (gdata, "sort_style");

	if ( sort_style > 3 )
		GTK_CLIST (Pan.article_ctree)->sort_type = GTK_SORT_DESCENDING;
	else
		GTK_CLIST (Pan.article_ctree)->sort_type = GTK_SORT_ASCENDING;

        gtk_clist_set_sort_column ( GTK_CLIST(Pan.article_ctree), sort_style % 4 );
}


void
articlelist_sort (void)
{
	GtkCList* tree = GTK_CLIST ( Pan.article_ctree );
        const int sort_column = tree->sort_column;

        switch ( sort_column )
        {
	case 0:
		gtk_clist_set_compare_func ( tree, articlelist_compare_subjects );
		break;

	case 1:
		gtk_clist_set_compare_func ( tree, articlelist_compare_commatized );
		break;

	case 4:
		gtk_clist_set_compare_func ( tree, articlelist_compare_text );
		break;

	default:
		gtk_clist_set_compare_func ( tree, NULL );
		break;
        }

	/* only sort the top-level articles -- it makes more sense to leave
	   threads sorted by thread
	   gtk_clist_sort ( tree ); */
	gtk_ctree_sort_node (
		GTK_CTREE(Pan.article_ctree),
		gtk_ctree_node_nth(GTK_CTREE(Pan.article_ctree),0) );
}

static int
atoi_commatized ( const gchar* str )
{
	int i = 0;
	for ( ; *str; ++str ) 
	{
		if ( *str==',' )
			continue;
		else if ( isdigit ( (int)*str ) )
			i = i*10 + (*str-'0');
		else break;
	}
	return i;
}

static gint
articlelist_compare_commatized (
	GtkCList *clist,
	gconstpointer ptr1,
	gconstpointer ptr2)
{
	register const char* text1 = GTK_CELL_TEXT(((GtkCListRow*)(ptr1))->cell[clist->sort_column])->text;
	register const char* text2 = GTK_CELL_TEXT(((GtkCListRow*)(ptr2))->cell[clist->sort_column])->text;

        if ( text1 && text2 )
		return atoi_commatized(text1) - atoi_commatized(text2);
	if ( !text2 )
		return (text1 != NULL);
	// !text1
		return -1;

}

static
gint
articlelist_compare_subjects (GtkCList* clist,
			      gconstpointer ptr1,
			      gconstpointer ptr2)
{
	register int val;
	register const char* text1 = GTK_CELL_TEXT(((GtkCListRow*)(ptr1))->cell[clist->sort_column])->text;
	register const char* text2 = GTK_CELL_TEXT(((GtkCListRow*)(ptr2))->cell[clist->sort_column])->text;

	if ( text1 && text2 ) {
		text1 = skip_reply_leader ( text1 );
		text2 = skip_reply_leader ( text2 );
		val = *text1 - *text2; /* for mass sorts, try to avoid overhead of function call */
		if ( !val )
			val = toupper(*text1) - toupper(*text2);
		if ( !val )
			val = g_strcasecmp ( text1+1, text2+1 );
		return val;
	}
	if ( !text2 )
		return (text1 != NULL);
	// !text1
		return -1;
}


static
gint
articlelist_compare_text (GtkCList* clist,
			  gconstpointer ptr1,
			  gconstpointer ptr2)
{
	register int val;
	register const char* text1 = GTK_CELL_TEXT(((GtkCListRow*)(ptr1))->cell[clist->sort_column])->text;
	register const char* text2 = GTK_CELL_TEXT(((GtkCListRow*)(ptr2))->cell[clist->sort_column])->text;

	if ( text1 && text2 ) {
		val = *text1 - *text2; /* for mass sorts, try to avoid overhead of function call */
		if ( !val )
			val = toupper(*text1) - toupper(*text2);
		if ( !val )
			val = g_strcasecmp ( text1+1, text2+1 );
		return val;
	}
	if ( !text2 )
		return (text1 != NULL);
	// !text1
		return -1;

}


static
void
articlelist_click_column (
	GtkCList* clist,
	int n )
{
	gint sort_style;

	if (clist->sort_column == n) { /* flip sort styles if the column was clicked twice */
		if ( clist->sort_type == GTK_SORT_ASCENDING ) {
			clist->sort_type = GTK_SORT_DESCENDING;
		}
		else {
			clist->sort_type =  GTK_SORT_ASCENDING;
		}
	}
	else
		clist->sort_type = GTK_SORT_ASCENDING;

	gtk_clist_set_sort_column (GTK_CLIST (clist), n);

        articlelist_sort ( );

	sort_style = n + (clist->sort_type == GTK_SORT_DESCENDING)*4;

	group_set_attrib_i (Pan.active_group, "sort_style", sort_style);
}


static void article_ctree_destroy_cb ( void )
{
	gtk_widget_destroy ( GTK_WIDGET ( articlelist_closed_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( articlelist_open_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( article_read_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( decode_queue_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( download_queue_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( error_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( binary_decoded_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( mp_complete_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( mp_incomplete_pixmap ) );
}


/*--------------------------------------------------------------------
 * generate the listing of articles, for the "Articles" tab
 *--------------------------------------------------------------------*/
GtkWidget*
create_articlelist_ctree (void)
{
	GtkWidget *s_window;

	char *titles[] = {
		_("Subject"),
		_("Lines"),
		_("Date"),
		_("Author"),
	};

	articlelist_closed_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(book_closed_xpm);
	articlelist_open_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(book_open_xpm);
	article_read_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(mini_page_xpm);
	decode_queue_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(bluecheck_xpm);
	download_queue_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(greencheck_xpm);
	error_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(x_xpm);
	binary_decoded_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_xpm);
	mp_complete_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_complete_xpm);
	mp_incomplete_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_incomplete_xpm);

	cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &unread_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &new_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &read_color))
		g_error ("couldn't allocate color");

	Pan.article_ctree = gtk_ctree_new_with_titles (4, 0, titles);
	gtk_ctree_set_line_style (GTK_CTREE(Pan.article_ctree),
				  GTK_CTREE_LINES_DOTTED);
	//GTK_CLIST_SET_FLAG(GTK_CLIST(Pan.article_ctree),CLIST_AUTO_RESIZE_BLOCKED);
	gtk_widget_show (Pan.article_ctree);

	s_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(s_window),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(s_window), Pan.article_ctree);

	gtk_clist_set_selection_mode (GTK_CLIST(Pan.article_ctree),
				      GTK_SELECTION_EXTENDED);

	gtk_clist_set_column_width (GTK_CLIST(Pan.article_ctree), 0, 400);
	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 1, TRUE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 2, TRUE);

	/* right justfiy lines and date columns */
	gtk_clist_set_column_justification (GTK_CLIST (Pan.article_ctree), 1, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification (GTK_CLIST (Pan.article_ctree), 2, GTK_JUSTIFY_RIGHT);

/* create the right click popup menu */
	article_ctree_menu = gnome_popup_menu_new (articlelist_menu_popup);

/* connect signals */
	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "button_press_event",
			    GTK_SIGNAL_FUNC(articlelist_button_press), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "click_column",
			    GTK_SIGNAL_FUNC(articlelist_click_column), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "destroy",
	                    GTK_SIGNAL_FUNC(article_ctree_destroy_cb), NULL);


	return s_window;
}
