/* Pan - Message Window general functions
 *
 * Copyright (C) 1999 Jason Leach
 * Copyright (C) 1999 Matt Eagleson (e.Messiah@superpimp.org)
 *
 * Authors: Jason Leach <leach@wam.umd.edu>
 *          Charles Kerr <charles@skywalker.ecn.ou.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 */

#include "pan.h"

static void update_info_pane (MessageWindow *mw, Message *msg);

static void message_window_save_ok_clicked (GtkWidget *widget, GtkWidget *file_entry);
static void message_window_save_cb (GtkWidget *widget, Message *msg);

static void message_window_cut_cb (GtkWidget *widget, Message *msg);
static void message_window_copy_cb (GtkWidget *widget, Message *msg);
static void message_window_paste_cb (GtkWidget *widget, Message *msg);

static void message_window_close (GtkWidget *widget, Message *msg);
static void message_window_appbar_toggle (GtkWidget *widget, Message *msg);

static GtkWidget* create_post_body_pane (MessageWindow *mw);
static GtkWidget* create_read_body_pane (GtkWidget *body_widget, Message *msg);

gboolean use_message_window_always = FALSE;

static GnomeUIInfo message_post_file_menu[] = {
//	GNOMEUIINFO_MENU_OPEN_ITEM (not_implemented, NULL),
	GNOMEUIINFO_ITEM_STOCK (N_("Send Now"), N_("Send this message now."), 
				message_send, GNOME_STOCK_PIXMAP_MAIL_SND),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (message_window_save_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PRINT_ITEM (not_implemented, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM (message_window_close, NULL),	
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_file_menu[] = {
//	GNOMEUIINFO_MENU_OPEN_ITEM (not_implemented, NULL),
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (message_window_save_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PRINT_ITEM (not_implemented, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM (message_window_close, NULL),	
	GNOMEUIINFO_END
};

static GnomeUIInfo message_edit_menu[] = {
	GNOMEUIINFO_MENU_CUT_ITEM (message_window_cut_cb, NULL),
	GNOMEUIINFO_MENU_COPY_ITEM (message_window_copy_cb, NULL),
	GNOMEUIINFO_MENU_PASTE_ITEM (message_window_paste_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_view_menu[] = {
	{
		GNOME_APP_UI_TOGGLEITEM,
		N_ ("Show Status Bar"), 0,
		message_window_appbar_toggle, 0, 0, 0, 0, 0, 0, 0
	},
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_message_menu[] = {
	{ GNOME_APP_UI_ITEM, N_("New"), N_("Send this message now."), message_post_window },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Reply to Sender"), N_("Send this message now."), message_reply_window },
	{ GNOME_APP_UI_ITEM, N_("Reply to Group"), N_("Send this message now."), message_reply_window },
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_message_menu[] = {
	{ GNOME_APP_UI_ITEM, N_("New"), N_("Create a new message."), message_post_window },
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_main_menu[] = {
	GNOMEUIINFO_MENU_FILE_TREE (message_read_file_menu),
	GNOMEUIINFO_MENU_EDIT_TREE (message_edit_menu),
	GNOMEUIINFO_SUBTREE (N_("_View"), &message_view_menu),
	GNOMEUIINFO_SUBTREE (N_("_Message"), &message_read_message_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_main_menu[] = {
	GNOMEUIINFO_MENU_FILE_TREE (message_post_file_menu),
	GNOMEUIINFO_MENU_EDIT_TREE (message_edit_menu),
	GNOMEUIINFO_SUBTREE (N_("_View"), &message_view_menu),
	GNOMEUIINFO_SUBTREE (N_("_Message"), &message_post_message_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_toolbar[] = {
	GNOMEUIINFO_ITEM_STOCK (N_("Reply"), N_("Post a Reply to this Message"),
				message_reply_window, GNOME_STOCK_PIXMAP_MAIL_RPL),
//	GNOMEUIINFO_ITEM_STOCK (N_("Reply Sender"), N_("Mail a Reply to the Author of this Message"),
//				message_reply_window, GNOME_STOCK_PIXMAP_MAIL_SND),
	GNOMEUIINFO_ITEM_STOCK (N_("Forward"), N_("Forward this Message"), 
				NULL, GNOME_STOCK_PIXMAP_MAIL_FWD),
	GNOMEUIINFO_SEPARATOR,
//	GNOMEUIINFO_ITEM_STOCK (N_("Spelling"), N_("Check spelling"),
//				NULL, GNOME_STOCK_PIXMAP_SPELLCHECK),
	GNOMEUIINFO_ITEM_STOCK (N_("Print"), N_("Print this message"),
				not_implemented, GNOME_STOCK_PIXMAP_PRINT),
	GNOMEUIINFO_SEPARATOR,
//	GNOMEUIINFO_ITEM_STOCK (N_("Previous"), N_("Read the previous message"),
//				message_previous, GNOME_STOCK_PIXMAP_UP),
//	GNOMEUIINFO_ITEM_STOCK (N_("Next"), N_("Read the next message"),
//				message_next, GNOME_STOCK_PIXMAP_DOWN),
//	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Close"), N_("Close this Window"),
				message_window_close, GNOME_STOCK_PIXMAP_CLOSE),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_toolbar[] = {
	GNOMEUIINFO_ITEM_STOCK (N_("Send"), N_("Send this message"), 
				message_send, GNOME_STOCK_PIXMAP_MAIL_SND),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Spelling"), N_("Check spelling"),
				NULL, GNOME_STOCK_PIXMAP_SPELLCHECK),
	GNOMEUIINFO_ITEM_STOCK (N_("Print"), N_("Print this message"),
				not_implemented, GNOME_STOCK_PIXMAP_PRINT),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Close"), N_("Close this Window without Sending Message"),
				message_window_close, GNOME_STOCK_PIXMAP_CLOSE),
	GNOMEUIINFO_END
};


static void
message_window_appbar_toggle (GtkWidget *widget, Message *msg)
{
	if ( !msg->window->appbar ) {
		msg->window->appbar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_USER );
		gnome_app_set_statusbar (GNOME_APP (msg->window->window), GTK_WIDGET (msg->window->appbar));
		if (msg->type == NNTP_ARTICLE_READ)
			gnome_app_install_menu_hints (GNOME_APP (msg->window->window), message_read_main_menu);
		else
			gnome_app_install_menu_hints (GNOME_APP (msg->window->window), message_post_main_menu);
	}
	else if (GTK_IS_WIDGET (msg->window->appbar->parent)) {
		gtk_widget_ref (msg->window->appbar);
		gtk_container_remove (GTK_CONTAINER (msg->window->appbar->parent), msg->window->appbar);
	}
	else {
		gtk_box_pack_start (GTK_BOX (GNOME_APP (msg->window->window)->vbox), msg->window->appbar, FALSE, FALSE, 0);
		gtk_widget_unref (msg->window->appbar);
	}
}


static void
message_window_save_ok_clicked (GtkWidget *widget, GtkWidget *file_entry)
{
	gchar *filename;
	FILE *msg_dump;
	gchar *header;
	gchar *body;
	gchar *msg_id;

	msg_id = gtk_object_get_data (GTK_OBJECT (file_entry), "msg_id");

	filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_entry));
	header = acache_load_header (msg_id);
	body = acache_load_body (msg_id);

	if ((msg_dump = fopen (filename, "w")) == NULL) {
		g_print ("fopen() failed to save message");
		return;
	}

	fprintf (msg_dump, "%s\n%s", header, body);
	fclose (msg_dump);

	g_free (header);
	g_free (body);

	gtk_widget_destroy (file_entry);
}


static void
message_window_save_cb (GtkWidget *widget, Message *msg)
{
	GtkWidget *file_entry = gtk_file_selection_new (_("Save message to file"));

	gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (file_entry));

	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (file_entry)->cancel_button),
				   "clicked",
				   GTK_SIGNAL_FUNC (gtk_widget_destroy),
				   GTK_OBJECT (file_entry));

	gtk_object_set_data (GTK_OBJECT (file_entry), "msg_id", msg->message_id);

	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_entry)->ok_button),
			    "clicked",
			    GTK_SIGNAL_FUNC (message_window_save_ok_clicked),
			    file_entry);

	gtk_widget_show_all (file_entry);
}


static void
message_window_open_cb (GtkWidget *widget, Message *msg)
{

}


static void
message_window_cut_cb (GtkWidget *widget, Message *msg)
{
	g_return_if_fail ( msg->window->body );

	gtk_editable_cut_clipboard (GTK_EDITABLE(msg->window->body));
}


static void
message_window_copy_cb (GtkWidget *widget, Message *msg)
{
	g_return_if_fail ( msg->window->body );

	gtk_editable_copy_clipboard (GTK_EDITABLE(msg->window->body));
}


static void
message_window_paste_cb (GtkWidget *widget, Message *msg)
{
	gtk_editable_paste_clipboard (GTK_EDITABLE(msg->window->body));
}


static void
message_window_destroy (MessageWindow *mw)
{
//	message_free (mw->message);
	g_free (mw->message);

	gtk_widget_destroy (mw->window);
}


static void
message_window_close (GtkWidget *widget, Message *msg)
{
//	gtk_widget_destroy (GTK_WIDGET (msg->window->window));
	message_window_destroy (msg->window);
}


/*---[ create_read_info_pane ]----------------------------------------
 * From:
 * Organization:
 * Reply-To: (optional)
 * Newsgroups:
 * Followup-To: (optional)
 * Date:
 * Subject:
 *--------------------------------------------------------------------*/
static void
populate_read_info_pane (MessageWindow *mw, GtkTable* table, const Message* msg)
{
	GtkWidget *label;
        int row = 0;

	/* from */
	label = gtk_label_new ( _("From:") );
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row+1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->from = gtk_label_new (msg->from);
	gtk_misc_set_alignment (GTK_MISC (mw->from), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), mw->from, 1, 2, row, row+1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	/* organization */
	++row;
	label = gtk_label_new ( _("Organization:") );
	gtk_misc_set_alignment( GTK_MISC (label), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row+1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	/* FIXME: should actually get the Organization: from the message,
	   not say None */
	mw->organization = gtk_label_new ("None");
	gtk_misc_set_alignment (GTK_MISC (mw->organization), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), mw->organization, 1, 2, row, row+1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

        /* reply-to */
        if ( !*msg->reply_to && !strcmp(msg->from,msg->reply_to) ) {
		++row;
		label = gtk_label_new ( _("Reply-To:") );
		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
		gtk_table_attach (GTK_TABLE(table), label, 0, 1, row,row+1,
				  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
		mw->reply_to = gtk_label_new (msg->reply_to);
		gtk_misc_set_alignment (GTK_MISC (mw->reply_to), 0.0, 0.0);
		gtk_table_attach (GTK_TABLE (table), mw->reply_to, 1, 2, row, row+1,
				  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);
	}

        /* newsgroups */
	++row;
	label = gtk_label_new ( _("Newsgroups:") );
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row+1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->newsgroups = gtk_label_new (msg->newsgroups);
	gtk_misc_set_alignment (GTK_MISC(mw->newsgroups), 0.0, 0.0);
	gtk_label_set_line_wrap (GTK_LABEL(mw->newsgroups), TRUE);
	gtk_table_attach (GTK_TABLE (table), mw->newsgroups, 1, 2, row, row+1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);


        /* followup-to */
        if ( !*msg->followup_to && !strcmp(msg->newsgroups,msg->followup_to) ) {
		++row;
		label = gtk_label_new ( _("Followup-To:") );
		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
		gtk_table_attach (GTK_TABLE(table), label, 0, 1, row,row+1,
				  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
		mw->followup_to = gtk_label_new (msg->followup_to);
		gtk_misc_set_alignment (GTK_MISC (mw->followup_to), 0.0, 0.5);
		gtk_table_attach (GTK_TABLE (table), mw->followup_to, 1, 2, row, row+1,
				  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);
	}

        /* date */
	++row;
	label = gtk_label_new ( _("Date:") );
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row+1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->date = gtk_label_new (msg->date);
	gtk_misc_set_alignment (GTK_MISC (mw->date), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), mw->date, 1, 2, row, row+1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

        /* subject */
	++row;
	label = gtk_label_new ( _("Subject:") );
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row+1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->subject = gtk_label_new (msg->subject);
	gtk_misc_set_alignment (GTK_MISC (mw->subject), 0.0, 0.0);
	gtk_label_set_line_wrap (GTK_LABEL(mw->subject), TRUE);
	gtk_table_attach (GTK_TABLE (table), mw->subject, 1, 2, row, row+1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	//	update_info_pane ( mw, msg );
}

static GtkWidget *
create_read_info_pane (MessageWindow *mw)
{
	GtkWidget *table;

	table = gtk_table_new (7, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	populate_read_info_pane ( mw, GTK_TABLE(table), mw->message );

	return table;
}


static void
update_title_with_subject (GtkWidget *widget, Message *msg)
{
	gtk_window_set_title (GTK_WINDOW (msg->window->window),
			      gtk_entry_get_text (GTK_ENTRY (widget)));
}


/*--------------------------------------------------------------------
 * Newsgroups:
 * CC:
 * Subject:
 *--------------------------------------------------------------------*/
static GtkWidget *
create_post_info_pane (MessageWindow *mw)
{
	GtkWidget *table;
	GtkWidget *label;
	gchar *p, *q, *r;

	table = gtk_table_new (4, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);

	label = gtk_label_new ( _("From:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->from = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(table), mw->from, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	if ((p = gnome_config_get_string ("/Pan/User/Email"))) {
		if ((q = gnome_config_get_string ("/Pan/User/Full_Name"))) {
			r = g_strdup_printf ("%s <%s>", q, p);
			g_free (q);
			g_free (p);
		} else {
			r = p;
		}
		gtk_entry_set_text (GTK_ENTRY(mw->from), r);
		g_free(r);
	}
	label = gtk_label_new (_("Organization:"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
			 GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->organization = gtk_entry_new();
	gtk_table_attach (GTK_TABLE(table), mw->organization, 1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	if ((p = gnome_config_get_string ("/Pan/User/Organization"))) {
		gtk_entry_set_text(GTK_ENTRY(mw->organization), p);
		g_free(p);
	}
	label = gtk_label_new ( _("Newsgroups:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 2, 3,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->newsgroups = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(table), mw->newsgroups, 1, 2, 2, 3,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	/* set the newsgroups entry field as best we can */
	p = NULL;
	if (mw->message && mw->message->subject) {
		if (mw->message->followup_to && *mw->message->followup_to)
			p = mw->message->followup_to;
		else if (mw->message->newsgroups && *mw->message->newsgroups)
			p = mw->message->newsgroups;
	}
	if ( !p && Pan.active_group )
		p = Pan.active_group->name;
	if ( p )
		gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), p );

	label = gtk_label_new ( _("Subject:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 3, 4,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->subject = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(table), mw->subject, 1, 2, 3, 4,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);
	if (mw->message && mw->message->subject) {
		if ( !g_strncasecmp ( "Re: ", mw->message->subject, 4 ) )
		{
			gtk_entry_set_text (GTK_ENTRY(mw->subject), mw->message->subject);
		}
		else {
			gchar* pch = g_strconcat ( "Re: ", mw->message->subject, NULL );
			gtk_entry_set_text( GTK_ENTRY(mw->subject), pch );
			g_free ( pch );
		}
	}

	gtk_signal_connect (GTK_OBJECT (mw->subject), "changed",
			    update_title_with_subject, mw->message);

	return table;
}


static GtkWidget*
create_post_body_pane (MessageWindow *mw)
{
	GtkWidget *scrolled_window;
	gchar* sig_file;

	/* setup color stuff */
	GdkColor quoted_col = { 0, 0, 0, 0xbbbb };
	GdkColormap *cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &quoted_col))
		g_error ("couldn't allocate color");

	if ( !mw->body )
		mw->body = gtk_text_new (NULL, NULL);

	gtk_text_set_word_wrap (GTK_TEXT (mw->body), TRUE);
	gtk_text_set_editable (GTK_TEXT (mw->body), TRUE);

	gtk_text_freeze (GTK_TEXT (mw->body));

	/* insert message body */
	if (mw->message->type == NNTP_ARTICLE_REPLY)
	{
		gchar *body, *tmp, *tmp2, c;
		gint sig_reached = 0;

		tmp = g_strdup_printf (_("On %s, %s wrote:\n"), mw->message->date, mw->message->from);
		gtk_text_insert (GTK_TEXT (mw->body), NULL, &mw->body->style->black, NULL, tmp, -1);
		g_free (tmp);

		body = g_strdup (mw->message->body);

		tmp = body;
		tmp2 = strchr (tmp, '\n');
		while ( tmp2 && !sig_reached ) {
			c = *(tmp2 + 1);
			*(tmp2 + 1) = '\0';
			if ( strstr (tmp, "-- ") ) {
				sig_reached = 1;
				continue;
			}
			gtk_text_insert (GTK_TEXT (mw->body), NULL, &quoted_col, NULL, "> ", 2);
			gtk_text_insert (GTK_TEXT (mw->body), NULL, &quoted_col, NULL, tmp, -1);
			*(tmp2 + 1) = c;
			tmp = tmp2 + 1;
			tmp2 = strchr (tmp, '\n');
		}
		g_free (body);

		gtk_text_insert (GTK_TEXT (mw->body), NULL, &mw->body->style->black, NULL, "\n\n", 1);
	}

	/* attach signature */
	sig_file = gnome_config_get_string ( "/Pan/User/Signature_File" );
	if ( sig_file && strlen(sig_file) )
	{
		GArray* sig = read_file ( sig_file );
		if ( sig != NULL )
		{
			gtk_text_insert (GTK_TEXT (mw->body), NULL, &mw->body->style->black, NULL, "\n-- \n\n", 5);
			gtk_text_insert (GTK_TEXT (mw->body), NULL, &mw->body->style->black, NULL, sig->data, sig->len);
			gtk_text_set_point (GTK_TEXT (mw->body), 0);
			g_array_free ( sig, TRUE );
		}
		g_free ( sig_file );
	}

	gtk_text_thaw (GTK_TEXT (mw->body));

	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (scrolled_window), mw->body);

	return scrolled_window;
}


/*---[ create_read_body_pane ]----------------------------------------
 * will be called from Reading code
 *--------------------------------------------------------------------*/
static GtkWidget *
create_read_body_pane (GtkWidget *body_widget, Message *msg)
{
	gchar* sig_file;

	/* setup color stuff */
	GdkColor quoted_col = { 0, 0, 0, 0xbbbb };
	GdkColormap *cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &quoted_col))
		g_error ("couldn't allocate color");


	if ( !body_widget )
		body_widget = gtk_text_new (NULL, NULL);

	gtk_text_set_word_wrap (GTK_TEXT (body_widget), TRUE);

	gtk_text_freeze (GTK_TEXT (body_widget));

	/* insert message body */
	if (msg->type == NNTP_ARTICLE_READ)
	{
		gchar *body, *tmp, *tmp2, c;

		body = g_strdup (msg->body);

		tmp = body;
		tmp2 = strchr (tmp, '\n');
		while (tmp2) {
			c = *(tmp2 + 1);
			*(tmp2 + 1) = '\0';
			if (*(tmp + 0) == '>')
				gtk_text_insert (GTK_TEXT (body_widget), NULL, &quoted_col, NULL, tmp, -1);
			else
				gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, tmp, -1);
			*(tmp2 + 1) = c;
			tmp = tmp2 + 1;
			tmp2 = strchr (tmp, '\n');
		}
		g_free (body);
	}

	if (msg->type == NNTP_ARTICLE_REPLY)
	{
		gchar *body, *tmp, *tmp2, c;
		gint sig_reached = 0;

		tmp = g_strdup_printf (_("On %s, %s wrote:\n"), msg->date, msg->from);
		gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, tmp, -1);
		g_free (tmp);

		body = g_strdup (msg->body);

		tmp = body;
		tmp2 = strchr (tmp, '\n');
		while ( tmp2 && !sig_reached ) {
			c = *(tmp2 + 1);
			*(tmp2 + 1) = '\0';
			if ( strstr (tmp, "-- ") ) {
				sig_reached = 1;
				continue;
			}
			gtk_text_insert (GTK_TEXT (body_widget), NULL, &quoted_col, NULL, "> ", 2);
			gtk_text_insert (GTK_TEXT (body_widget), NULL, &quoted_col, NULL, tmp, -1);
			*(tmp2 + 1) = c;
			tmp = tmp2 + 1;
			tmp2 = strchr (tmp, '\n');
		}
		g_free (body);

		gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, "\n", 1);
	}

	/* attach signature */
	if (msg->type != NNTP_ARTICLE_READ)
	{
		sig_file = gnome_config_get_string ( "/Pan/User/Signature_File" );
		if ( sig_file && strlen(sig_file) )
		{
			GArray* sig = read_file ( sig_file );
			if ( sig != NULL )
			{
				gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, "\n-- \n\n", 5);
				gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, sig->data, sig->len);
				gtk_text_set_point (GTK_TEXT (body_widget), 0);
				g_array_free ( sig, TRUE );
			}
			g_free ( sig_file );
		}
	}

	gtk_text_thaw (GTK_TEXT (body_widget));

//	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
//	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
//					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
//	gtk_container_add (GTK_CONTAINER (scrolled_window), body_widget);

//	return scrolled_window;

	gtk_widget_show (body_widget);

	return body_widget;
}


void
update_info_pane (MessageWindow *mw, Message *msg)
{
	//GList* l;
#if 0
	/* rebuild the children.
	   I'm doing it this way because I can't get items to not draw in a GtkTable,
	   and we don't want to show followup-to or reply-to if they're not set.
	   <sigh>  -- csk */
	while ( (( l = gtk_container_children (GTK_CONTAINER(mw->read_info_pane)) )) != NULL )
		gtk_container_remove ( GTK_CONTAINER(mw->read_info_pane), GTK_WIDGET(l->data) );
	populate_read_info_pane (mw, GTK_TABLE(mw->read_info_pane), msg);
#else
	gtk_label_set_text (GTK_LABEL (mw->from), msg->from);
	gtk_label_set_text (GTK_LABEL (mw->organization), "None");
	gtk_label_set_text (GTK_LABEL (mw->date), msg->date);
	gtk_label_set_text (GTK_LABEL (mw->newsgroups), msg->newsgroups);
	gtk_label_set_text (GTK_LABEL (mw->subject), msg->subject);
        gtk_label_set_text (GTK_LABEL (mw->followup_to), msg->followup_to);
        gtk_label_set_text (GTK_LABEL (mw->reply_to), msg->reply_to);
#endif
	gtk_window_set_title (GTK_WINDOW (mw->window), msg->subject);

}


/*---[ update_body_pane ]---------------------------------------------
 * will only be called for messages being "read"
 *--------------------------------------------------------------------*/
void
update_body_pane (GtkWidget *body_widget, gchar *body)
{
	gint i;
	gchar *tmp, *tmp2, c;
	
	/* setup colors */
	GdkColor quoted_col = { 0, 0, 0, 0xbbbb };
	GdkColormap *cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &quoted_col))
		g_error ("couldn't allocate color");

	gtk_text_freeze (GTK_TEXT (body_widget));

	/* clear old text */
	i = gtk_text_get_length (GTK_TEXT (body_widget));
	if (i > 0) {
		gtk_text_set_point (GTK_TEXT (body_widget), i);
		gtk_text_backward_delete (GTK_TEXT (body_widget), i);
	}

	/* insert new text (colored quoted) */
	tmp = body;
	tmp2 = strchr (tmp, '\n');
	while (tmp2) {
		c = *(tmp2 + 1);
		*(tmp2 + 1) = '\0';
		if (*(tmp + 0) == '>')
			gtk_text_insert (GTK_TEXT (body_widget), NULL, &quoted_col, NULL, tmp, -1);
		else
			gtk_text_insert (GTK_TEXT (body_widget), NULL, &body_widget->style->black, NULL, tmp, -1);
		*(tmp2 + 1) = c;
		tmp = tmp2 + 1;
		tmp2 = strchr (tmp, '\n');
	}

	gtk_text_set_point (GTK_TEXT (body_widget), 0);

	gtk_text_thaw (GTK_TEXT (body_widget));
}


static void
message_window_update (MessageWindow *mw, Message *msg)
{
	if (mw && mw->body)
	{
		update_info_pane (mw, msg);
		update_body_pane (mw->body, msg->body);
	}
	else
	{
		update_body_pane (text, msg->body);
	}
}


void
message_post_window (GtkWidget *widget, Message *msg)
{
	if ( !msg || msg->window )
		msg = message_new (NULL);
	
	msg->type = NNTP_ARTICLE_POST;
	message_window_new (msg);
}


static void
message_post_window_create (MessageWindow *mw)
{
	GtkWidget *vbox;

	mw->window = gnome_app_new ("Post", "New Message");
	vbox = gtk_vbox_new (FALSE, 0);

	gnome_app_set_contents (GNOME_APP (mw->window), vbox);

	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	
	gnome_app_create_menus_with_data (GNOME_APP (mw->window), message_post_main_menu, mw->message);
	gnome_app_create_toolbar_with_data (GNOME_APP (mw->window), message_post_toolbar, mw->message);

	gtk_box_pack_start (GTK_BOX(vbox), create_post_info_pane (mw), FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX(vbox), create_post_body_pane (mw), TRUE, TRUE, 0);

//	{
//		GtkWidget *scrolled_window = gtk_scrolled_window_new (NULL, NULL);
//		GtkWidget *body_widget = create_body_pane (mw->body, mw->message);
//		create_body_pane (NULL, mw->message);

//		gtk_text_set_editable (GTK_TEXT (mw->body), TRUE);
//
//		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
//					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
//		gtk_container_add (GTK_CONTAINER (scrolled_window), mw->body);
//		gtk_box_pack_start (GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);
//	}

	gtk_widget_show_all (mw->window);
}


void
message_read_window (GtkWidget *widget, Message *msg)
{
	if ( !msg ) {
		const char *message_id = articlelist_get_selected_message_id ();
		text_set_from_message_id ( message_id, NNTP_ARTICLE_READ, TRUE );
		return;
	}

	msg->type = NNTP_ARTICLE_READ;
	message_window_new (msg);
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static void
message_read_window_create (MessageWindow *mw)
{
	GtkWidget *vbox;

	/* init vars */
	mw->window = gnome_app_new ("Read", mw->message->subject);
//	mw->appbar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_USER);
	vbox = gtk_vbox_new (FALSE, 0);

	gnome_app_set_contents (GNOME_APP (mw->window), vbox);
//	gnome_app_set_statusbar (GNOME_APP (mw->window), GTK_WIDGET (mw->appbar));

	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	
	/* create menubar and toolbar */
	gnome_app_create_menus_with_data (GNOME_APP (mw->window),
					  message_read_main_menu,
					  mw->message);
	gnome_app_create_toolbar_with_data (GNOME_APP (mw->window),
					    message_read_toolbar,
					    mw->message);
//	gnome_app_install_menu_hints (GNOME_APP (mw->window), message_main_menu);

//	((GtkCheckMenuItem *) (message_view_menu[0].widget))->active = 1;

	/* pack widgets */
	mw->read_info_pane = create_read_info_pane (mw);
	gtk_box_pack_start (GTK_BOX (vbox), mw->read_info_pane, FALSE, FALSE, GNOME_PAD_SMALL);
//	gtk_box_pack_start (GTK_BOX (vbox), create_read_info_pane (mw), FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), gtk_hseparator_new(), FALSE, FALSE, 2);
	{
		GtkWidget *scrolled_window = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_container_add (GTK_CONTAINER (scrolled_window), create_read_body_pane (NULL, mw->message));
		gtk_box_pack_start (GTK_BOX (vbox), scrolled_window, TRUE, TRUE, 0);
	}
		
	/* display window */
	gtk_widget_show_all (mw->window);
}


void
message_popup_reply_window (GtkWidget *widget, Message *msg)
{
	if ( !msg && article ) {
		msg = message_new (article);

		msg->type = NNTP_ARTICLE_REPLY;
		message_window_new (message_copy (msg));
	}
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
void
message_reply_window (GtkWidget *widget, Message *msg)
{
	if ( !msg ) {
		const char *message_id = articlelist_get_selected_message_id ();
		text_set_from_message_id ( message_id, NNTP_ARTICLE_REPLY, TRUE );
		return;
	}

	msg->type = NNTP_ARTICLE_REPLY;
	message_window_new (message_copy (msg));
}


static void
message_reply_window_create (MessageWindow *mw)
{
	GtkWidget *vbox;
	gchar* pch;

	if ( !g_strncasecmp ( "Re: ", mw->message->subject, 4 ) )
	{
		pch = g_strdup (mw->message->subject);
	}
	else
		pch = g_strconcat ( "Re: ", mw->message->subject, NULL );

	mw->window = gnome_app_new ("Reply", pch);
	vbox = gtk_vbox_new (FALSE, 0);

	gnome_app_set_contents (GNOME_APP (mw->window), vbox);

	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	
	gnome_app_create_menus_with_data (GNOME_APP (mw->window), message_post_main_menu, mw->message);
	gnome_app_create_toolbar_with_data (GNOME_APP (mw->window), message_post_toolbar, mw->message);

	gtk_box_pack_start (GTK_BOX(vbox), create_post_info_pane (mw), FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX(vbox), create_post_body_pane (mw), TRUE, TRUE, 0);

//	{
//		GtkWidget *scrolled_window = gtk_scrolled_window_new (NULL, NULL);
//		GtkWidget *body_widget = create_body_pane (mw->body, mw->message);
//		mw->body = create_body_pane (mw->body, mw->message->body);

//		gtk_text_set_editable (GTK_TEXT (mw->body), TRUE);
		
//		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
//					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
//		gtk_container_add (GTK_CONTAINER (scrolled_window), mw->body);
//		gtk_box_pack_start (GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);
//	}

	gtk_widget_show_all (mw->window);

	g_free (pch);
}


void
message_window_new (Message *msg)
{
	MessageWindow *mw = g_malloc0 (sizeof(MessageWindow));

	mw->message = msg;
	msg->window = mw;

	switch (msg->type) {
	case NNTP_ARTICLE_READ:
		message_read_window_create (mw);
		break;
	case NNTP_ARTICLE_POST:
		message_post_window_create (mw);
		break;
	case NNTP_ARTICLE_REPLY:
		message_reply_window_create (mw);
		break;
//	case EMAIL_REPLY:
	case EMAIL_FORWARD:
	default:
		break;
	}
}
