/* Pan - In-window text widget for viewing messages
 *
 * Copyright (C) 1999 Matt Eagleson (e.Messiah@superpimp.org)
 *
 * Authors: Matt Eagleson <e.Messiah@superpimp.org>
 *          Jason Leach <leach@wam.umd.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 */

#include "pan.h"


/***
 *** global variables
 ***/
GtkWidget *text;
GtkWidget *text_rightclick_popup;

article_data *article;

static GnomeUIInfo text_menu_popup[] = {
	{ GNOME_APP_UI_ITEM, N_("Reply to Sender"), N_("Create a reply to the sender."), message_popup_reply_window },
	{ GNOME_APP_UI_ITEM, N_("Reply to _Group"), N_("Post a reply to the message on the news server."), message_popup_reply_window },
	{ GNOME_APP_UI_ITEM, N_("Forward"), N_("Forward this message to someone."), not_implemented },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Show Full Header"), N_("Show the full header for the current message."), not_implemented },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Select All"), N_("Select all messages"), text_select_all,
	  NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 'A', GDK_CONTROL_MASK, NULL},
	GNOMEUIINFO_END
};

static gchar welcome[] =
{
"Pan 0.6.6 \n"
"Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org) \n"
"\n"
"The Pan source code is Copyright (C) 1999 by \n"
"        Matt Eagleson (e.Messiah@superpimp.org) \n"
"        Charles Kerr (charles@skywalker.ecn.ou.edu) \n"
"        Jason Leach (leach@wam.umd.edu) \n"
"\n"
"Pan homepage: http://www.superpimp.org/ \n"
"\n"
"This program is free software; you can redistribute it and/or modify \n"
"it under the terms of the GNU General Public License as published by \n"
"the Free Software Foundation; either version 2 of the License, or \n"
"(at your option) any later version. \n"
"\n"
"This program is distributed in the hope that it will be useful, \n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of \n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \n"
"GNU General Public License for more details. \n"
"\n"
"You should have received a copy of the GNU General Public License \n"
"along with this program; if not, write to the Free Software \n"
"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA"
};


/***
 *** Private Function Prototypes
 ***/
static void realize_text_cb (GtkWidget *text, gpointer data);
static void text_button_press (GtkWidget *widget, GdkEventButton *bevent);


/***
 *** Public Routines
 ***/

void
text_select_all (void)
{
	gtk_editable_select_region (GTK_EDITABLE (text), 0, -1);
}


/*--------------------------------------------------------------------
 * called when the 'realize' signal is sent to the text widget,
 * not sure what it really does though... add comments ???
 *--------------------------------------------------------------------*/
static void
realize_text_cb (GtkWidget *text, gpointer data)
{
	GtkAdjustment* hadj = GTK_ADJUSTMENT(GTK_TEXT(text)->hadj);
	GtkAdjustment* vadj = GTK_ADJUSTMENT(GTK_TEXT(text)->vadj);
	text_clear ();
	gtk_text_insert (GTK_TEXT(text), NULL, &text->style->black, NULL, welcome, -1);
	gtk_text_set_point ( GTK_TEXT(text), 0 );

	gtk_adjustment_set_value ( GTK_ADJUSTMENT(vadj), 0 );
	gtk_text_set_adjustments ( GTK_TEXT(text), hadj, vadj );
}


/*---[ text_clear ]---------------------------------------------------
 * wipe the text widget clean
 *--------------------------------------------------------------------*/
void
text_clear (void)
{
	gint i;

	pan_lock();

	gui_set_headers_default (NULL, NULL);
	i = gtk_text_get_length (GTK_TEXT(text));
	if (i > 0) {
		gtk_text_set_point (GTK_TEXT(text), i);
		gtk_text_backward_delete (GTK_TEXT(text), i);
	}
	gtk_widget_queue_draw (GTK_WIDGET (text));

	pan_unlock ();
}


/*---[ text_button_press ]--------------------------------------------
 * callback for when a mouse button is pressed inside the text widget
 *--------------------------------------------------------------------*/
static void
text_button_press (GtkWidget *widget, GdkEventButton *bevent)
{
	gtk_widget_grab_focus (GTK_WIDGET (text));

	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
	
	if (bevent->button == 3) {
		const int sel = (article) != NULL;

		GList* l = GTK_MENU_SHELL(text_rightclick_popup)->children; /* l is "reply to sender" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is "reply to group" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is "forward" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is separator */
		l = l->next; /* l is "select all" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );

		gtk_menu_popup (GTK_MENU(text_rightclick_popup), NULL, NULL, NULL,
				NULL, bevent->button, bevent->time);
	}
}


/*---[ text_create ]--------------------------------------------------
 * create a scrolled text area that displays a message's body
 *--------------------------------------------------------------------*/
GtkWidget *
text_create (void)
{
	GtkWidget *scrolled_window;

	text = gtk_text_new (NULL, NULL);

	widget_set_font ( GTK_WIDGET(text), message_body_font );

	gtk_text_set_editable (GTK_TEXT(text), FALSE);
	gtk_text_set_word_wrap (GTK_TEXT(text), TRUE);

	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	/* Add a handler to put a message in the text widget when it is realized */
	gtk_signal_connect (GTK_OBJECT(text), "realize",
			    GTK_SIGNAL_FUNC (realize_text_cb), NULL);

	gtk_signal_connect (GTK_OBJECT(text), "button_press_event",
			    GTK_SIGNAL_FUNC (text_button_press), NULL);

	gtk_container_add (GTK_CONTAINER (scrolled_window), text);

	text_rightclick_popup = gnome_popup_menu_new (text_menu_popup);

	return scrolled_window;
}


void
text_update (Message *msg)
{
	char *body;
	article_data *adata = NULL;
	article_db db;

	g_free ( Pan.active_message_id );
	Pan.active_message_id = g_strdup ( msg->message_id );

	body = acache_load_body ( msg->message_id );

	db = ahdb_ref ( Pan.active_server, Pan.active_group );
	adata = ahdb_get ( db, msg->message_id );
	ahdb_unref ( db );

	if ( msg && body )
	{
		if ( use_message_window_always || (msg->needs_window) )
		{
			message_import_adata (msg, adata);
			
			pan_lock ();
			message_window_new (msg);
			pan_unlock ();
		}
		else
		{
			pan_lock ();

			gui_set_headers_default ( NULL, adata );
			update_body_pane (text, body);

			pan_unlock ();

			g_free (article);
			article = article_copy (adata);
		}
	}

	/* mark it read here, because damnit, it's just been drawn
	   up for the user's viewing - JEL */
	if (adata->state != STATE_READ) {
		group_add_attrib_i (Pan.active_group, "Read", 1);
		article_add_flag (adata, STATE_READ, TRUE);
	}
}


void
text_set_from_message_id (const gchar *message_id,
			  MessageWindowType type,
			  gboolean needs_window)
{
	Message *msg = message_new (NULL);
	msg->message_id = g_strdup (message_id);
	msg->type = type;
	msg->needs_window = needs_window;

	if (acache_hit_body (message_id) == 0)
		text_update (msg);
	else
		queue_add_article_download (Pan.active_server,
					    Pan.active_group,
					    msg,
					    QUEUE_FOREGROUND);

//	message_free (msg);
}


#if 0
void
text_update (const gchar *message_id)
{
	char* body;
	article_data* adata = NULL;
	article_db db;

	/* update active message */
	if ( Pan.active_message_id )
		g_free ( Pan.active_message_id );
	Pan.active_message_id = g_strdup ( message_id );

	/* FIXME: Rethink an option to show full headers
	 * headers = gnome_config_get_int ("/Pan/Message/Headers=1"); */
	body = acache_load_body ( message_id );

	/* get the header information */
	db = ahdb_ref ( Pan.active_server, Pan.active_group );
	adata = ahdb_get ( db, message_id );
	ahdb_unref ( db );

	if (adata && body)
	{
		if ( use_message_window_always ) {
			Message *msg = message_new (adata);
			msg->type = NNTP_ARTICLE_READ;
		
			pan_lock ();
			message_read_window (NULL, msg);
			pan_unlock ();

			message_free (msg);
		}
		else {
//			text_clear ();
			pan_lock ();

			gui_set_headers_default ( NULL, adata );
			update_body_pane (text, body);

			pan_unlock ();

			g_free (article);
			article = article_copy (adata);
		}

		g_free (adata);
		g_free (body);
/*
		text_clear ();
		pan_lock ();
	
		g_free (article);
		article = article_copy (adata);

		gtk_text_freeze (GTK_TEXT (text));

		gtk_label_set_text (GTK_LABEL (from_label), adata->author);
		gtk_label_set_text (GTK_LABEL (subject_label), adata->subject);

		gtk_text_insert (GTK_TEXT(text), NULL, &text->style->black, NULL, body, -1);

		gtk_text_thaw (GTK_TEXT(text));
		gtk_widget_queue_draw (GTK_WIDGET (text));

		pan_unlock ();

		g_free (adata);
		g_free (body);
*/
	}
}


article_data *
displayed_article (void)
{
	return article;
}
#endif
